@@ -1,14 +1,9 @@
 from __future__ import with_statement
 
-from fabric.api import *
-from fabric.contrib.project import rsync_project
-from fabric.version import get_version
+from fabric.api import lcd, local, task
 
 
-docs_host = 'jforcier@fabfile.org'
-
-
-@task
+@task(default=True)
 def build(clean='no', browse_='no'):
     """
     Generate the Sphinx documentation.
@@ -29,14 +24,3 @@ def browse():
     Open the current dev docs in a browser tab.
     """
     local("open docs/_build/html/index.html")
-
-
-@task
-@hosts(docs_host)
-def push():
-    """
-    Build docs and zip for upload to RTD
-    """
-    build(clean='yes')
-    v = get_version('short')
-    local("cd docs/_build/html && zip -r ../%s.zip ." % v)
